#version 150 core

in vec2 textureCoord;

uniform sampler2D   albedoBuffer;
uniform sampler2D   normalBuffer;
uniform sampler2D   lightBuffer;
uniform sampler2D   depthBuffer;
uniform sampler2D   ssaoBuffer;
uniform samplerCube shadowCubemapBufferMask;

uniform mat4 projectionMatrix;

uniform int   drawMode;
uniform float gammaCorrection;
uniform float exposureLevel;
uniform bool  drawLights;
uniform bool  drawSSAO;
uniform vec3  ambientIntensity;

uniform float lightRadius;
uniform vec3  softShadowsMaskMipLevels;

uniform bool homogeneousReversedDepth;

out vec4 color;

vec3 screenGamma = vec3( 1.0f / gammaCorrection );

float linearizeDepth( in float z ) {
  float A     = projectionMatrix[ 2 ].z;
  float B     = projectionMatrix[ 3 ].z;
  float zFar  = -B / ( 1.0 - A );
  float zNear = B / ( 1.0 + A );

  if ( homogeneousReversedDepth ) {
    return zNear / ( zNear - z * ( zNear - zFar ) );
  }
  return 2.0 * zNear / ( zFar + zNear - z * ( zFar - zNear ) );
}

vec3 getLighting() {
  vec3  light            = textureLod( lightBuffer, textureCoord, 0 ).rgb;
  float ssao             = drawSSAO ? textureLod( ssaoBuffer, textureCoord, 0 ).x : 1.0f;
  vec3  selfIllumination = vec3( textureLod( normalBuffer, textureCoord, 0 ).w );
  return max( light + ambientIntensity * ssao, selfIllumination );
}

// Exposure tone mapping
vec3 getHDR( in vec3 colorValue ) {
  vec3 mapped = vec3( 1.0f ) - exp( -colorValue * exposureLevel );
  return pow( mapped, screenGamma );
}

float linearizeDepth( in float z, in float zNear, in float zFar ) {
  if ( homogeneousReversedDepth ) {
    return zFar * zNear / ( zNear - z * ( zNear - zFar ) );
  }
  return zFar * 2.0 * zNear / ( zFar + zNear - z * ( zFar - zNear ) );
}

void main() {
  if ( drawMode == 1 ) {
    vec3 colorSample = textureLod( albedoBuffer, textureCoord, 0 ).rgb;
    if ( drawLights ) {
      colorSample *= getLighting();
    }
    color = vec4( getHDR( colorSample ), 1.0f );
  } else if ( drawMode == 2 ) {
    color = vec4( textureLod( normalBuffer, textureCoord, 0 ).xyz, 1.0f );
  } else if ( drawMode == 3 ) {
    color = vec4( getHDR( getLighting() ), 1.0f );
  } else if ( drawMode == 4 ) {
    vec3 ssaoValue = vec3( textureLod( ssaoBuffer, textureCoord, 0 ).x );
    color          = vec4( ssaoValue, 1.0f );
  } else if ( drawMode == 5 ) {
    color = vec4( vec3( textureLod( depthBuffer, textureCoord, 0 ).x ), 1.0f );
  } else if ( drawMode == 6 ) {
    color = vec4( getHDR( textureLod( albedoBuffer, textureCoord, 0 ).rgb ), 1.0f );
  } else if ( drawMode == 7 ) {
    float phi   = textureCoord.x * 3.1415 * 2;
    float theta = ( -textureCoord.y + 0.5 ) * 3.1415;

    vec3  dir = vec3( cos( phi ) * cos( theta ), sin( theta ), sin( phi ) * cos( theta ) );
    float depth =
      linearizeDepth( textureLod( shadowCubemapBufferMask, dir, softShadowsMaskMipLevels[ 2 ] ).x, 0.1, lightRadius ) / lightRadius;
    color = vec4( vec3( depth ), 1 );
  } else if ( drawMode == 9 ) {
    ivec2 bufferSize = textureSize( depthBuffer, 0 );

    float offset = 4.0 / bufferSize.x;

    vec2 offsets[ 9 ] = vec2[](  //
      vec2( -offset, offset ),   // top-left
      vec2( 0.0f, offset ),      // top-center
      vec2( offset, offset ),    // top-right
      vec2( -offset, 0.0f ),     // center-left
      vec2( 0.0f, 0.0f ),        // center-center
      vec2( offset, 0.0f ),      // center-right
      vec2( -offset, -offset ),  // bottom-left
      vec2( 0.0f, -offset ),     // bottom-center
      vec2( offset, -offset )    // bottom-right
    );

    float kernel[ 9 ] = float[]( -1, -1, -1, -1, 8, -1, -1, -1, -1 );

    float sampleTex[ 9 ];
    for ( int i = 0; i < 9; ++i ) {
      sampleTex[ i ] = texture( depthBuffer, textureCoord.xy + offsets[ i ] ).x;
    }
    float col = 0.0f;
    for ( int i = 0; i < 9; ++i ) {
      col += sampleTex[ i ] * kernel[ i ];
    }

    if ( textureCoord.x <= offset || textureCoord.y <= offset || textureCoord.x + offset > 1 || textureCoord.y + offset > 1 ) {
      color = vec4( vec3( sampleTex[ 4 ] ), 1.0 );
      return;
    }

    color = vec4( vec3( col > 0.00025f ? 1.0f : 0.0f ), 1.0 );
  }
}
