#version 150 core

layout( std140 ) uniform Matrices {
  mat4 viewProjectionMatrix;
  mat4 viewProjectionInversedMatrix;
};

uniform sampler2D   depthBuffer;
uniform sampler2D   normalBuffer;
uniform samplerCube lightDepthCubemapBuffer;
uniform samplerCube lightDepthCubemapMask;
uniform sampler2D   lightDepthBuffer;
uniform sampler2D   lightDepthMask;

uniform mat4 lightProjectionMatrix;
uniform vec2 screenSize;

// для восстановления позиции из глубины
uniform bool  homogeneousReversedDepth;
uniform bool  drawLights;
uniform bool  drawShadows;
uniform bool  useSoftShadows;
uniform int   softShadowsSamplesCount;
uniform vec3  softShadowsMaskMipLevels;
uniform float lightSize;
uniform bool  useContactShadows;

uniform float projectedRadius;
uniform float negInvRadius2;

uniform bool  isNearestLight;
uniform int   lightType;
uniform vec3  wsLightPosition;
uniform vec3  lightColor;
uniform float lightRadius;
uniform float lightAttenuationLinear;
uniform float lightAttenuationQuadratic;
uniform vec3  wsLightDirection;
uniform float lightConeInner;
uniform float lightConeOuter;

uniform vec3 wsCameraPosition;
uniform vec3 wsCameraDirection;

out vec3 fragmentColor;

vec3 ssToNDC( in vec3 ssPosition ) {
  vec3 ndcPosition = vec3( ssPosition.xy * 2.0f - 1.0f, ssPosition.z );
  if ( !homogeneousReversedDepth ) {
    // если глубина должна быть в диапазоне [-1, 1]
    ndcPosition.z = ndcPosition.z * 2.0f - 1.0f;
  }
  return ndcPosition;
}
vec3 ndcToSS( in vec3 ndcPosition ) {
  vec3 ssPosition = vec3( ndcPosition.xy * 0.5f + 0.5f, ndcPosition.z );
  if ( !homogeneousReversedDepth ) {
    // если глубина была в диапазоне [-1, 1]
    ssPosition.z = ssPosition.z * 0.5f + 0.5f;
  }
  return ssPosition;
}

// вычисляем позицию в мировых координатах на основе однородных координат [0, 1] экрана и глубины
vec3 ssToWSPosition( in vec3 ssPosition ) {
  vec4 wsPosition = viewProjectionInversedMatrix * vec4( ssToNDC( ssPosition ), 1.0f );
  return wsPosition.xyz / wsPosition.w;
}
vec3 getWSPositionFromDepth( in vec2 textureCoords ) {
  float z = textureLod( depthBuffer, textureCoords, 0 ).x;
  return ssToWSPosition( vec3( textureCoords, z ) );
}

vec3 wsToSSPosition( in vec3 wsPosition ) {
  vec4 csPosition = viewProjectionMatrix * vec4( wsPosition, 1.0f );
  if ( csPosition.w < 0.0f ) {
    csPosition.w = -csPosition.w;
  }
  vec3 ndcPosition = csPosition.xyz;
  if ( csPosition.w != 0.0f ) {
    ndcPosition /= csPosition.w;
  }
  return ndcToSS( ndcPosition );
}

// проверяем, лежит ли точка за пределами видимой области
bool isClipped( in vec3 ndcPosition ) {
  if ( ndcPosition.x < -1.0f || ndcPosition.x > 1.0f || ndcPosition.y < -1.0f || ndcPosition.y > 1.0f ||
    ( homogeneousReversedDepth && ndcPosition.z < 0.0f ) || ( !homogeneousReversedDepth && ndcPosition.z < -1.0f ) ||
    ndcPosition.z > 1.0f ) {
    return true;
  }
  return false;
}

// Let V be the light-to-fragment direction vector.
// V needs to be in world space since the shadow cube map is normally axis aligned in world space
// thus the "max(abs(V.x),abs(V.y),abs(V.z))"-part only works if V is a world space direction vector.
// The Z value in the respective cube side frustum/"eye space" can be calculated by taking the max of the absolute values of V's components.
// Z = max(abs(V.x),abs(V.y),abs(V.z))
// Then, to be precise, we should negate Z because in OpenGL, the negative Z-axis points into the screen/view frustum.
// Now we want to get the depth buffer "compatible" value of that -Z.
// Looking at the OpenGL perspective matrix we see that, for any homogeneous vector multiplied with that matrix,
// the resulting z value is completely independent of the vector's x and y components.
// So we can simply multiply this matrix with the homogeneous vector (0,0,-Z,1) and we get the vector (components):
// x = 0
// y = 0
// z = (-Z * -(f+n) / (f-n)) + (-2*f*n / (f-n))
// w = Z
// Then we need to do the perspective divide, so we divide z by w (Z) which gives us:
// z' = (f+n) / (f-n) - 2*f*n / (Z* (f-n))
// This z' is in OpenGL's normalized device coordinate (NDC) range [-1,1] and needs to be transformed into a depth buffer compatible range of [0,1]:
// z_depth_buffer_compatible = (z' + 1.0) * 0.5
float vectorToDepth( in vec3 fragmentToLight ) {
  vec3  absVector = abs( fragmentToLight );
  float lsDepth   = max( absVector.x, max( absVector.y, absVector.z ) );
  // clip space
  vec4 csPosition = lightProjectionMatrix * vec4( 0.0f, 0.0f, -lsDepth, 1.0f );
  // NDC depth
  float ndcDepth = csPosition.z / csPosition.w;

  if ( !homogeneousReversedDepth ) {
    // если clip space в диапазоне [-1, 1], преобразуем в [0, 1]
    return ndcDepth * 0.5f + 0.5f;
  }
  return ndcDepth;
}

vec3 getLightSpacePosition( in vec3 wsPosition ) {
  // clip space
  vec4 lsPosition = lightProjectionMatrix * vec4( wsPosition, 1.0f );
  // NDC
  vec3 projCoords = lsPosition.xyz / lsPosition.w;
  projCoords.xy   = projCoords.xy * 0.5f + 0.5f;
  if ( !homogeneousReversedDepth ) {
    // если clip space в диапазоне [-1, 1], преобразуем в [0, 1]
    projCoords.z = projCoords.z * 0.5f + 0.5f;
  }
  return projCoords;
}

// Квадратичное затухание
float quadraticAttenuation( in float lightDistance, in float lightRadius ) {
  // расчитываем затухание только в пределах радиуса
  float d         = min( lightRadius, lightDistance );
  float nominator = clamp( 1.0f - pow( d / lightRadius, 4 ), 0.0f, 1.0f );
  // делимое в квадрате для более плавного затухания
  nominator *= nominator;
  // квадратичное затухание с параметрами
  float denominator = 1.0f + lightAttenuationLinear * d + lightAttenuationQuadratic * d * d;
  return nominator / denominator;
}

float lowPrecisionHash( vec2 p ) {
  vec3 p3 = fract( vec3( p.xyx ) * 0.13f );
  p3 += dot( p3, p3.yzx + 3.333f );
  return fract( ( p3.x + p3.y ) * p3.z ) * 20.1f;
}

float linearizeDepth( in float z, in float zNear, in float zFar ) {
  if ( homogeneousReversedDepth ) {
    return zFar * zNear / ( zNear - z * ( zNear - zFar ) );
  }
  return zFar * 2.0 * zNear / ( zFar + zNear - z * ( zFar - zNear ) );
}

vec2 vogelDiskSample( int sampleIndex, float spinAngle ) {
  const float goldenAngle = 2.39996f;
  float       theta       = sampleIndex * goldenAngle + spinAngle;
  return vec2( cos( theta ), sin( theta ) );
}

float getSoftShadowSample( in int sampleIndex, in int maxSampleCount, in float spinAngle, in mat3 tbnMatrix, in float occluderCoefficient,
  in vec3 wsPosition, out float distanceToOccluder ) {
  float ssRadius   = sqrt( ( sampleIndex + 0.5f ) / maxSampleCount );
  vec2  unitOffset = vogelDiskSample( sampleIndex, spinAngle );

  vec3 tsSamplePosition = occluderCoefficient * ssRadius * vec3( unitOffset, ssRadius );
  vec3 wsSamplePosition = tbnMatrix * tsSamplePosition;

  float sampleFragmentDepth;
  float sampleDepth;
  if ( lightType == 1 ) {
    // cubemap
    vec3 sampleDirection = wsLightPosition - ( wsPosition + wsSamplePosition );
    sampleFragmentDepth  = linearizeDepth( vectorToDepth( sampleDirection ), 0.1f, lightRadius );
    sampleDepth          = linearizeDepth( texture( lightDepthCubemapBuffer, -sampleDirection ).x, 0.1f, lightRadius );
  } else {
    vec3 sampleLightSpacePosition = getLightSpacePosition( wsPosition + wsSamplePosition );
    sampleFragmentDepth           = linearizeDepth( sampleLightSpacePosition.z, 0.1f, lightRadius );
    sampleDepth                   = linearizeDepth( texture( lightDepthBuffer, sampleLightSpacePosition.xy ).x, 0.1f, lightRadius );
  }

  distanceToOccluder = sampleFragmentDepth - sampleDepth;

  return sampleDepth > sampleFragmentDepth ? 1.0f : 0.0f;
}

float calculateShadow(
  in vec3 wsPosition, in vec3 wsNormal, in vec3 wsFragToLight, in float lightDistance, out float penumbraCoefficient ) {
  float fragmentDepth;
  float depth;

  // устанавливаем mip-уровень маски полутени в зависимости от расстояния до источника,
  // чтобы не покрывать слишком большие области избыточными выборками
  float maskMipLevel = softShadowsMaskMipLevels[ 0 ];
  if ( lightDistance >= 10 && lightDistance < 30 ) {
    maskMipLevel = softShadowsMaskMipLevels[ 1 ];
  } else if ( lightDistance >= 30 ) {
    maskMipLevel = softShadowsMaskMipLevels[ 2 ];
  }

  float minDiskRadius = 24.0f;  // будет поделено на разрешение теневой карты, чтобы получить соответствующий радиус размытия: ~0.1f для 256

  if ( lightType == 1 ) {
    // cubemap
    fragmentDepth       = linearizeDepth( vectorToDepth( wsFragToLight ), 0.1f, lightRadius );
    depth               = linearizeDepth( texture( lightDepthCubemapBuffer, -wsFragToLight ).x, 0.1f, lightRadius );
    penumbraCoefficient = useSoftShadows ? textureLod( lightDepthCubemapMask, -wsFragToLight, maskMipLevel ).x : 0.0f;
    minDiskRadius /= textureSize( lightDepthCubemapBuffer, 0 ).x;
  } else {
    vec3 lightSpacePosition = getLightSpacePosition( wsPosition );
    fragmentDepth           = linearizeDepth( lightSpacePosition.z, 0.1f, lightRadius );
    depth                   = linearizeDepth( texture( lightDepthBuffer, lightSpacePosition.xy ).x, 0.1f, lightRadius );
    penumbraCoefficient     = useSoftShadows ? textureLod( lightDepthMask, lightSpacePosition.xy, maskMipLevel ).x : 0.0f;
    minDiskRadius /= textureSize( lightDepthBuffer, 0 ).x;
  }

  // расстояние от камеры до фрагмента сцены
  float       ssDiskRadius = projectedRadius / length( wsPosition - wsCameraPosition );
  const float MIN_RADIUS   = 8.0f;  // pixels
  if ( !useSoftShadows || penumbraCoefficient == 0.0f || ssDiskRadius <= MIN_RADIUS ) {
    return depth > fragmentDepth ? 1.0f : 0.0f;
  }

  // меняющийся для каждого фрагмента угол вращения диска
  float randomPatternRotationAngle = lowPrecisionHash( gl_FragCoord.xy );

  // матрица перевода диска из tangent-space в world-space, используем wsFragToLight как случайный вектор
  vec3 tangent   = normalize( wsFragToLight - wsNormal * dot( wsFragToLight, wsNormal ) );
  mat3 tbnMatrix = mat3( tangent, cross( wsNormal, tangent ), wsNormal );

  float shadowCoefficient = 0.0f;

  float maxDiskRadius = max( minDiskRadius, lightSize );

  float fragmentDistanceToOccluder = fragmentDepth - depth;
  float lightSizeCoefficient       = lightSize * 0.1f;
  float defaultOccluderCoefficient = max( fragmentDistanceToOccluder * lightSizeCoefficient, minDiskRadius );

  // если стартуем в тени, используем коэффициент с учетом полутени, если нет - средний радиус
  float occluderCoefficient = fragmentDistanceToOccluder > 0.0f ? defaultOccluderCoefficient : mix( maxDiskRadius, minDiskRadius, 0.25f );

  for ( int i = 0; i < softShadowsSamplesCount; ++i ) {
    float distanceToOccluder;
    shadowCoefficient += getSoftShadowSample(
      i, softShadowsSamplesCount, randomPatternRotationAngle, tbnMatrix, occluderCoefficient, wsPosition, distanceToOccluder );

    if ( distanceToOccluder > 0.0f ) {
      // если сэмпл в тени - усредняем радиус согласно данным сэмпла
      occluderCoefficient = mix( occluderCoefficient, max( distanceToOccluder * lightSizeCoefficient, minDiskRadius ), 0.5f );
    }
  }
  return shadowCoefficient / softShadowsSamplesCount;
}

// Approximated soft shadows, based on https://iquilezles.org/articles/rmshadows and https://www.shadertoy.com/view/tscSRS and further fixed and improved
float contactSoftShadow( vec3 wsPosition, vec3 wsFragToLightDir ) {
  float stepSize  = 0.005;
  float tmax      = 10.0;
  float lightSize = 0.1;

  float t   = stepSize;
  float res = 1.0f;
  for ( int i = 0; i < 256; i++ ) {
    vec3 wsCurrentPosition = wsPosition + t * wsFragToLightDir;

    vec4 csPosition = viewProjectionMatrix * vec4( wsCurrentPosition, 1.0f );
    // если вышли за область видимости, прерываем обход
    if ( csPosition.w < 0.0f ) {
      break;
    }
    vec3 ndcPosition = csPosition.xyz;
    if ( csPosition.w != 0.0f ) {
      ndcPosition /= csPosition.w;
    }
    // if ( isClipped( ndcPosition ) ) {
    //  diffuse *= vec3(1,0,0);
    // 	break;
    // }
    vec3 ssCurrentPosition = ndcToSS( ndcPosition );

    float sampleHeight = textureLod( depthBuffer, ssCurrentPosition.xy, 0 ).x;

    // if ( ( !homogeneousReversedDepth && sampleHeight > ssCurrentPosition.z ) || ( homogeneousReversedDepth && sampleHeight < ssCurrentPosition.z ) ) {
    // 	// фрагмент освещен
    // 	return 1.0f;
    // }

    vec3 wsSamplePosition = ssToWSPosition( vec3( ssCurrentPosition.xy, sampleHeight ) );

    float bias = 0.0004f;
    if ( !homogeneousReversedDepth ) {
      bias /= 1.0f - ssCurrentPosition.z;
    } else {
      bias /= ssCurrentPosition.z;
    }
    float h = length( wsCurrentPosition - wsSamplePosition );
    res     = min( res, h / ( lightSize * t ) );
    // приращение шага
    t += stepSize;
    //if ( t > tmax || h < bias ) {
    //	break;
    //}
  }
  res = max( res, 0.0f );
  return res;
  //return 0.25*(1.0+res)*(1.0+res)*(2.0-res); // smoothstep
}

vec3 calculateDiffuse( in vec3 wsNormal, in vec3 wsFragToLightDir ) {
  // рассеянный свет
  float diffuseCoeff = max( dot( wsNormal, wsFragToLightDir ), 0.0f );
  vec3  diffuse      = diffuseCoeff * lightColor;

  if ( !drawLights ) {
    return diffuse;
  }
  if ( lightType == 2 ) {
    // если spot
    float theta     = dot( wsFragToLightDir, normalize( -wsLightDirection ) );
    float epsilon   = lightConeInner - lightConeOuter;
    float intensity = clamp( ( theta - lightConeOuter ) / epsilon, 0.0f, 1.0f );
    diffuse *= intensity;
  } else if ( lightType == 3 ) {
    // если directional
    diffuse *= max( dot( wsFragToLightDir, normalize( -wsLightDirection ) ), 0.0f );
  }

  return diffuse;
}

void main() {
  vec2  textureCoords    = gl_FragCoord.xy / screenSize;
  vec3  wsPosition       = getWSPositionFromDepth( textureCoords );
  vec3  wsFragToLight    = wsLightPosition - wsPosition;
  float lightDistance    = length( wsFragToLight );
  vec3  wsFragToLightDir = normalize( wsFragToLight );

  // считаем только те фрагменты, которые находятся на освещаемой источником поверхности
  if ( lightDistance > lightRadius || ( lightType == 2 && dot( wsFragToLightDir, -wsLightDirection ) < lightConeOuter ) ) {
    discard;
  }

  vec3 wsNormal = textureLod( normalBuffer, textureCoords, 0 ).xyz;

  // и повернуты к нему лицом
  float fragToLightCoefficient = dot( wsNormal, wsFragToLightDir );
  if ( fragToLightCoefficient <= 0.0f ) {
    discard;
  }

  vec3  diffuse     = calculateDiffuse( wsNormal, wsFragToLightDir );
  float attenuation = lightType == 3 ? 1.0f : quadraticAttenuation( lightDistance, lightRadius );

  fragmentColor = diffuse * attenuation;

  if ( !drawShadows ) {
    return;
  }

  float penumbraCoefficient;
  float shadowCoefficient = calculateShadow( wsPosition, wsNormal, wsFragToLight, lightDistance, penumbraCoefficient );

  // if ( isNearestLight && penumbraCoefficient > 0.0f ) {
  //   fragmentColor = vec3( 1, 0, 0 ) * shadowCoefficient;
  //  return;
  // }

  fragmentColor *= shadowCoefficient;
}
